#!/usr/bin/env python
# encoding: utf-8


class HttpException(Exception):
    def __init__(self, request, response):
        self.__request = request
        self.__response = response

        status = response.get_status() if response else 500
        message = response.get_error()

        Exception.__init__(self, status, message)

    def get_request(self):
        return self.__request

    def get_response(self):
        return self.__response


#!/usr/bin/env python
# encoding: utf-8
import json
import urllib
import httplib
from urlparse import urlparse

from .http_exception import HttpException
from .headers import *
from .response import Response


POST = 'POST'
GET = 'GET'
DELETE = 'DELETE'
PUT = 'PUT'

ALLOWED_METHODS = [GET, POST, PUT, DELETE]


class Request(Headers):
    def __init__(self, method, url, query_params=None, body=None, headers=None):
        Headers.__init__(self)

        if method not in ALLOWED_METHODS:
            raise Exception('Unknown method')

        self.__method = method
        self.__url = url
        self.__query_params = query_params if query_params else {}
        self.__body = body

        self.set_headers({
            ACCEPT: JSON_CONTENT_TYPE,
            CONTENT_TYPE: JSON_CONTENT_TYPE
        })

        if headers:
            self.set_headers(headers)

    def get_url(self):
        return self.__url

    def send(self):

        url = urlparse(self.get_url())
        if url.scheme == "https":
            conn = httplib.HTTPSConnection(url.hostname, url.port)
        else:
            conn = httplib.HTTPConnection(url.hostname, url.port)
        try:
            conn.request(self.get_method(),
                         self.get_url(),
                         body=self.get_encoded_body(),
                         headers=self.get_headers())

            response = conn.getresponse()
            body = response.read()
            status_code = response.status
            headers = dict(response.getheaders())

            response = Response(status_code, body, dict(headers))
            if not response.check_status():
                raise HttpException(self, response)

            return response

        finally:
            conn.close()

    def get_url_with_query_string(self):
        url = self.__url
        query = urllib.urlencode(self.__query_params)
        if query:
            url = ('&' if self.__url.find('?') > 0 else '?') + query
        return url

    def get_encoded_body(self):
        if self.is_json():
            return json.dumps(self.__body)
        elif self.is_url_encoded():
            return urllib.urlencode(self.__body)
        else:
            return self.__body

    def is_put(self):
        return self.__method == PUT

    def is_get(self):
        return self.__method == GET

    def is_post(self):
        return self.__method == POST

    def is_delete(self):
        return self.__method == DELETE

    def get_method(self):
        return self.__method

    def set_method(self, method):
        self.__method = method

    def set_url(self, url):
        self.__url = url

    def get_body(self):
        return self.__body

    def set_body(self, body):
        self.__body = body

    def get_query_params(self):
        return self.__query_params

    # FIXME
    def is_loaded(self):
        return False


#!/usr/bin/env python
# encoding: utf-8

import base64
from .auth import Auth
from ..http.request import *
from ..http.response import *


ACCOUNT_ID = '~'
ACCOUNT_PREFIX = '/account/'
URL_PREFIX = '/restapi'
TOKEN_ENDPOINT = '/restapi/oauth/token'
API_VERSION = 'v1.0'
ACCESS_TOKEN_TTL = 3600  # 60 minutes
REFRESH_TOKEN_TTL = 36000  # 10 hours
REFRESH_TOKEN_TTL_REMEMBER = 604800  # 1 week


class Platform:
    def __init__(self, key, secret, server):
        self.__server = server
        self.appKey = key
        self.appSecret = secret
        self.__auth = Auth()
        self.__account = ACCOUNT_ID

    def set_auth_data(self, auth_data=None):
        self.__auth.set_data(auth_data)
        return self

    def get_auth_data(self):
        return self.__auth.get_data()

    def is_authorized(self, refresh=True):
        if not self.__auth.is_access_token_valid() and refresh:
            self.refresh()
        if not self.__auth.is_access_token_valid():
            raise Exception('Access token is not valid after refresh timeout')

    def authorize(self, user_name, extension, password, remember=False):
        response = self.auth_call(POST, TOKEN_ENDPOINT, body={
            'grant_type': 'password',
            'username': user_name,
            'extension': extension,
            'password': password,
            'access_toket_ttl': ACCESS_TOKEN_TTL,
            'refresh_token_ttl': REFRESH_TOKEN_TTL_REMEMBER if remember else REFRESH_TOKEN_TTL
        })
        self.__auth.set_data(response.get_json(False))
        self.__auth.set_remember(remember)

    def refresh(self):
        if not self.__auth.is_refresh_token_valid():
            raise Exception('Refresh token has expired')

        response = self.auth_call(POST, TOKEN_ENDPOINT, body={
            'grant_type': 'refresh_token',
            'refresh_token': self.__auth.get_refresh_token(),
            'access_token_ttl': ACCESS_TOKEN_TTL,
            'refresh_token_ttl': REFRESH_TOKEN_TTL_REMEMBER if self.__auth.is_remember() else REFRESH_TOKEN_TTL
        })

        self.__auth.set_data(response.get_json(False))

        return response

    def logout(self):
        response = self.auth_call(POST, TOKEN_ENDPOINT + '/revoke', body={
            'token': self.__auth.get_access_token
        })
        self.__auth.reset()
        return response

    def api_call(self, method, url, query_params=None, body=None, headers=None):
        self.is_authorized()
        request = Request(method, url, query_params, body, headers)
        request.set_header(AUTHORIZATION, self.__get_auth_header())
        request.set_url(self.api_url(request.get_url(), {'addServer': True}))
        return request.send()

    def auth_call(self, method, url, query_params=None, body=None, headers=None):
        request = Request(method, url, query_params, body, headers)
        request.set_header(AUTHORIZATION, 'Basic ' + self.__get_api_key())
        request.set_header(CONTENT_TYPE, URL_ENCODED_CONTENT_TYPE)
        request.set_url(self.api_url(request.get_url(), {'addServer': True}))
        request.set_method(POST)
        return request.send()

    def __get_api_key(self):
        return base64.b64encode(self.appKey + ':' + self.appSecret)

    def __get_auth_header(self):
        return self.__auth.get_token_type() + ' ' + self.__auth.get_access_token()

    def api_url(self, url, options=None):
        built_url = ''
        options = options if options else {}

        if 'addServer' in options and options['addServer'] and url.find('http://') < 0 and url.find('https://') < 0:
            built_url += self.__server

        if url.find(URL_PREFIX) < 0:
            built_url += URL_PREFIX + '/' + API_VERSION

        if url.find(ACCOUNT_PREFIX) < 0:
            built_url = built_url.replace(ACCOUNT_PREFIX + ACCOUNT_ID, ACCOUNT_PREFIX + self.__account)

        built_url += url

        if 'addMethod' in options and options['addMethod']:
            built_url += ('&' if url.find('?') >= 0 else '?') + '_method=' + options['addMethod']

        if 'addToken' in options and options['addToken']:
            built_url += ('&' if url.find('?') >= 0 else '?') + 'access_token=' + options['addMethod']

        return built_url

    def get(self, url, query_params=None, headers=None):
        return self.api_call(GET, url, query_params, None, headers)

    def post(self, url, query_params=None, body=None, headers=None):
        return self.api_call(POST, url, query_params, body, headers)

    def put(self, url, query_params=None, body=None, headers=None):
        return self.api_call(PUT, url, query_params, body, headers)

    def delete(self, url, query_params=None, body=None, headers=None):
        return self.api_call(DELETE, url, query_params, body, headers)





#!/usr/bin/env python
# encoding: utf-8

HEADER_SEPARATOR = ':'
CONTENT_TYPE = 'content-type'
AUTHORIZATION = 'authorization'
ACCEPT = 'accept'
URL_ENCODED_CONTENT_TYPE = 'application/x-www-form-urlencoded'
JSON_CONTENT_TYPE = 'application/json'
MULTIPART_CONTENT_TYPE = 'multipart/mixed'


class Headers:
    def __init__(self):
        self.headers = {}

    def set_header(self, key, val):
        self.headers[key.lower()] = val

    def set_headers(self, headers):
        for k in headers.keys():
            self.set_header(k, headers[k])

    def get_header(self, key):
        return self.headers.get(key.lower())

    def has_header(self, key):
        return key.lower() in self.headers

    def get_headers(self):
        return self.headers

    def get_headers_array(self):
        return [k.lower() + HEADER_SEPARATOR + str(self.headers[k]) for k in self.headers.keys()]

    def get_content_type(self):
        return self.get_header(CONTENT_TYPE)

    def set_content_type(self, ct):
        self.set_header(CONTENT_TYPE, ct)

    def is_content_type(self, ct):
        return self.get_content_type().find(ct) >= 0

    def is_json(self):
        return self.is_content_type(JSON_CONTENT_TYPE)

    def is_multipart(self):
        return self.is_content_type(MULTIPART_CONTENT_TYPE)

    def is_url_encoded(self):
        return self.is_content_type(URL_ENCODED_CONTENT_TYPE)



#!/usr/bin/env python
# encoding: utf-8

from time import time

RELEASE_TIMEOUT = 10


class Auth:
    def __init__(self):
        self.__remember = False

        self.__token_type = ''

        self.__access_token = ''
        self.__expires_in = 0
        self.__expire_time = 0

        self.__refresh_token = ''
        self.__refresh_token_expires_in = 0
        self.__refresh_token_expire_time = 0

        self.__scope = ''
        self.__owner_id = ''

    def set_data(self, auth_data=None):

        if not auth_data:
            return self

        # Misc

        if 'remeber' in auth_data:
            self.__remember = auth_data.get('remember')

        if 'token_type' in auth_data:
            self.__token_type = auth_data.get('token_type')

        if 'owner_id' in auth_data:
            self.__owner_id = auth_data.get('owner_id')

        if 'scope' in auth_data:
            self.__scope = auth_data.get('scope')

        # Access Token

        if 'access_token' in auth_data:
            self.__access_token = auth_data.get('access_token')

        if 'expires_in' in auth_data:
            self.__expires_in = auth_data.get('expires_in')

        if 'expire_time' not in auth_data and 'expires_in' in auth_data:
            self.__expire_time = time() + auth_data.get('expires_in')
        elif 'expire_time' in auth_data:
            self.__expire_time = auth_data.get('expire_time')

        # Refresh Token

        if 'refresh_token' in auth_data:
            self.__refresh_token = auth_data.get('refresh_token')

        if 'refresh_token_expires_in' in auth_data:
            self.__refresh_token_expires_in = auth_data.get('refresh_token_expires_in')

        if 'refresh_token_expire_time' not in auth_data and 'refresh_token_expires_in' in auth_data:
            self.__refresh_token_expire_time = time() + auth_data.get('refresh_token_expires_in')
        elif 'expire_time' in auth_data:
            self.__refresh_token_expire_time = auth_data.get('refresh_token_expire_time')

        return self

    def get_data(self):
        return {
            'remember': self.__remember,
            'token_type': self.__token_type,

            'access_token': self.__access_token,
            'expires_in': self.__expires_in,
            'expire_time': self.__expire_time,

            'refresh_token': self.__refresh_token,
            'refresh_token_expires_in': self.__refresh_token_expires_in,
            'refresh_token_expire_time': self.__refresh_token_expire_time,

            'scope': self.__scope,
            'owner_id': self.__owner_id
        }

    def reset(self):
        self.__remember = False

        self.__token_type = ''

        self.__access_token = ''
        self.__expires_in = 0
        self.__expire_time = 0

        self.__refresh_token = ''
        self.__refresh_token_expires_in = 0
        self.__refresh_token_expire_time = 0

        self.__scope = ''
        self.__owner_id = ''

    def get_access_token(self):
        return self.__access_token

    def get_refresh_token(self):
        return self.__refresh_token

    def get_token_type(self):
        return self.__token_type

    def is_access_token_valid(self):
        return self.__is_token_date_valid(self.get_data().get('expire_time'))

    def is_refresh_token_valid(self):
        return self.__is_token_date_valid(self.get_data().get('refresh_token_expire_time'))

    def set_remember(self, val):
        self.set_data({'remember': True if val else False})

    def is_remember(self):
        return self.get_data().get('remember', False)

    @staticmethod
    def __is_token_date_valid(token_date):
        return token_date > time()


#!/usr/bin/env python
# encoding: utf-8
import base64
from threading import Timer

from ..http.request import *


EVENTS = {
    'notification': 'notification',
    'removeSuccess': 'removeSuccess',
    'removeError': 'removeError',
    'renewSuccess': 'renewSuccess',
    'renewError': 'renewError',
    'subscribeSuccess': 'subscribeSuccess',
    'subscribeError': 'subscribeError'
}

RENEW_HANDICAP = 60


class Subscription:
    def __init__(self, platform):
        self.__platform = platform
        self.__event_filters = []
        self.__timeout = None
        self.__subscription = {
            'eventFilters': [],
            'expirationTime': '',  # 2014-03-12T19:54:35.613Z
            'expiresIn': 0,
            'deliveryMode': {
                'transportType': 'PubNub',
                'encryption': False,
                'address': '',
                'subscriberKey': '',
                'secretKey': ''
            },
            'id': '',
            'creationTime': '',  # 2014-03-12T19:54:35.613Z
            'status': '',  # Active
            'uri': ''
        }
        self.__pubnub = None
        self.__listeners = {}

    def on(self, event, fn):
        if event not in self.__listeners:
            self.__listeners[event] = []
        self.__listeners[event].append(fn)

    def __emit(self, event, data=None):
        if event in self.__listeners:
            for l in self.__listeners[event]:
                l(data)

    def register(self, options=None):
        if self.is_subscribed():
            self.renew(options)
        else:
            self.subscribe(options)

    def add_events(self, events):
        self.__event_filters += events
        pass

    def set_events(self, events):
        self.__event_filters = events

    def subscribe(self, options=None):
        options = options if options else {}
        if 'events' in options:
            self.__event_filters = options['events']
        try:
            if not self.__event_filters or len(self.__event_filters) == 0:
                raise Exception('Events are undefined')
            body = {
                'eventFilters': self.__get_full_events_filter(),
                'deliveryMode': {
                    'transportType': 'PubNub'
                }
            }
            response = self.__platform.api_call(POST, '/restapi/v1.0/subscription', body=body)
            data = response.get_json(False)
            self.__update_subscription(data)
            self.__subscribe_at_pubnub()
            self.__emit(EVENTS['subscribeSuccess'], data)
        except Exception as e:
            self.un_subscribe()
            self.__emit(EVENTS['subscribeError'], e)
            raise e

    def renew(self, options=None):
        options = options if options else {}
        self.__event_filters = options['events'] if 'events' in options else []
        self.__clear_timeout()
        try:
            if not self.__subscription or ('id' not in self.__subscription) or not self.__subscription['id']:
                raise Exception('Subscription ID is required')
            if not self.__event_filters or len(self.__event_filters) == 0:
                raise Exception('Events are undefined')
            body = {
                'eventFilters': self.__get_full_events_filter()
            }
            response = self.__platform.api_call(PUT, '/restapi/v1.0/subscription' + self.__subscription['id'], body=body)
            data = response.get_data(False)
            self.__update_subscription(data)
            self.__emit(EVENTS['subscribeSuccess'], data)
        except Exception as e:
            self.un_subscribe()
            self.__emit(EVENTS['renewError'], e)

    def remove(self):
        try:
            if not self.__subscription or ('id' not in self.__subscription) or not self.__subscription['id']:
                raise Exception('Subscription ID is required')
            self.__platform.api_call(DELETE, '/restapi/v1.0/subscription' + self.__subscription['id'])
            self.un_subscribe()
            self.__emit(EVENTS['removeSuccess'])
        except Exception as e:
            self.un_subscribe()
            self.__emit(EVENTS['removeError'], e)

    def destroy(self):
        self.un_subscribe()

    def un_subscribe(self):
        self.__clear_timeout()
        self.__un_subscribe_at_pubnub()
        self.__subscription = None

    def is_subscribed(self):
        s = self.__subscription
        return ('deliveryMode' in s and s['deliveryMode']) and \
               ('subscriberKey' in s['deliveryMode'] and s['deliveryMode']['subscriberKey']) and \
               ('address' in s['deliveryMode'] and s['deliveryMode']['address'])

    def __update_subscription(self, data):
        self.__clear_timeout()
        self.__subscription = data
        self.__set_timeout()

    def __subscribe_at_pubnub(self):
        from Pubnub import Pubnub, AES

        if not self.is_subscribed():
            return
        # TODO check this stuff
        s_key = self.__subscription['deliveryMode']['subscriberKey']
        self.__pubnub = Pubnub(subscribe_key=s_key, ssl_on=False, publish_key='')

        def callback(message, channel):
            if self.is_subscribed() and ('encryptionKey' in self.__subscription['deliveryMode']):
                key = base64.b64decode(self.__subscription['deliveryMode']['encryptionKey'])
                data = base64.b64decode(message)
                obj2 = AES.new(key)
                decrypted = obj2.decrypt(data)
                self.__notify(decrypted)
            else:
                self.__notify(message)

        def error(message):
            print("ERROR : " + str(message))

        def connect(message):
            print("CONNECTED")

        def reconnect(message):
            print("RECONNECTED")

        def disconnect(message):
            print("DISCONNECTED")

        self.__pubnub.subscribe(self.__subscription['deliveryMode']['address'], callback=callback, error=error,
                                connect=connect, reconnect=reconnect, disconnect=disconnect)

    def __notify(self, message):
        self.__emit(EVENTS['notification'], message)

    def __un_subscribe_at_pubnub(self):
        if self.__pubnub and self.is_subscribed():
            self.__pubnub.unsubscribe(self.__subscription['deliveryMode']['address'])

    def __get_full_events_filter(self):
        return [self.__platform.api_url(e) for e in self.__event_filters]

    def __set_timeout(self):
        time_to_expiration = self.__subscription['expiresIn'] - RENEW_HANDICAP
        self.__timeout = Timer(time_to_expiration, self.renew)
        self.__timeout.start()

    def __clear_timeout(self):
        if self.__timeout:
            self.__timeout.cancel()


if __name__ == '__main__':
    pass


# !/usr/bin/env python
# encoding: utf-8
import json
from email.feedparser import FeedParser

from .headers import *


UNAUTHORIZED_STATUS = 401


class Response(Headers):
    def __init__(self, status, body, headers=None):  # TODO Add statusText
        """
         Builds response by parsing raw response on-the-wire. Should be refactored out. Parsing should not happen here.
         It's not the place for this.
        :param status:
        :param raw:
        :param headers:
        :return:
        """
        Headers.__init__(self)

        self.__body = body
        self.__status = status

        if status is None:
            raise Exception('Empty status was received')

        if isinstance(headers, dict):
            self.set_headers(headers)

    def check_status(self):
        """
        Returns True if HTTP status is 200 OK
        :return:
        """
        return 200 <= self.__status < 300

    def get_body(self):
        """
        Returns whole response body as a string.
        :return:
        """
        return self.__body

    def get_json(self, as_object=True):
        """
        Returns:
         - JSON for application/json response,
         - or raises exception
        """
        if not self.is_json():
            raise Exception('Response is not JSON')

        parsed = json.loads(self.__body)
        return parsed if not as_object else unfold(parsed)

    def get_status(self):
        """
        Returns HTTP status code
        :return:
        """
        return self.__status

    def get_responses(self):
        """
        Returns:
         - sub-responses, if this response is multipart
         - or raises exception
        :return:
        """
        if not self.is_multipart():
            raise Exception('Response is not Batch (Multipart)')

        parts = self.__break_into_parts()
        if len(parts) < 1:
            # sic! not specific extension
            raise Exception("Malformed Batch Response (not enough parts)")
        # TODO Defensive checking for part[0] content type
        # TODO JSON parsing errors handling
        statuses = json.loads(parts[0].get_payload())
        if len(statuses["response"]) != len(parts) - 1:
            raise Exception("Malformed Batch Response (not-consistent number of parts)")

        responses = []

        for response, payload in zip(statuses["response"], parts[1:]):
            responses.append(Response(response["status"], payload.get_payload(), dict(payload)))

        return responses

    def get_error(self):
        message = self.__status + ' Unknown error'  # TODO Use statusText
        data = self.get_json(False)

        if 'message' in data:
            message = data['message']
        elif 'error_description' in data:
            message = data['error_description']
        elif 'description' in data:
            message = data['description']

        return message

    def __break_into_parts(self):

        p = FeedParser()
        for h in self.get_headers_array():
            p.feed(h + "\r\n")
        p.feed("\r\n")
        p.feed(self.__body)
        msg = p.close()
        parts = msg.get_payload()
        return parts


PYTHON_KEYWORDS = (
    "and", "del", "from", "not", "while", "as", "elif", "global", "or", "with", "assert", "else", "if", "pass", "yield",
    "break", "except", "import", "rint", "class", "exec", "in", "raise", "continue", "finally", "is", "return", "def",
    "for", "lambda", "try",)


class JsonObject:
    def __init__(self):
        pass


def safe_name(n):
    if n in PYTHON_KEYWORDS:
        return n + "_"
    else:
        return n


def unfold(d):
    if isinstance(d, dict):
        o = JsonObject()
        for k, v in d.iteritems():
            o.__dict__[safe_name(k)] = unfold(v)
        return o
    elif isinstance(d, list):
        o = [unfold(x) for x in d]
        return o
    else:
        return d


from distutils.core import setup

VERSION = __import__("rcsdk").VERSION

setup(
    name='rcsdk',
    packages=[
        'rcsdk',
        'rcsdk.http',
        'rcsdk.platform',
        'rcsdk.subscription'
    ],
    version=VERSION,
    description='RingPlatform Python SDK',
    author="RingCentral, Inc.",
    author_email="devsupport@ringcentral.com",
    url="https://github.com/ringcentral/python-sdk",
    download_url="https://github.com/ringcentral/python-sdk/tarball/%s" % VERSION,
    keywords=['sdk', 'ringcentral', 'ringplatform'],
    install_requires=[
        'pubnub>=3.7.0',
    ],
    classifiers=[],
)


#!/usr/bin/env python
# encoding: utf-8

VERSION = '0.3.0'


from .platform import Platform
from .subscription import Subscription


class RCSDK:
    def __init__(self, key, secret, server):
        self.__platform = Platform(key, secret, server)

    def get_platform(self):
        return self.__platform

    def get_subscription(self):
        return Subscription(self.__platform)
